/*
 * backstack.c
 *
 * Author: J. Steingraeber
 * Date: 2003-04-12
 *
 * Released to the Public Domain
 */

#include "modeldef.h"
#include "backstack.h"

typedef struct tagBACKSTACK 
{
  FILE_INFO fi;
  byte submode;
} BACKSTACK;

#define MAXSTACK 100

BACKSTACK backstack[MAXSTACK];

int backstackpushptr = 0;
int backstackpopptr = 0;

int nextPtr(int p)
{
  if (p == (MAXSTACK - 1))
  {
    return 0;
  }

  return (p + 1);
}

int prevPtr(int p)
{
  if (p == 0)
  {
    return (MAXSTACK - 1);
  }

  return (p - 1);
}

int wouldDuplicateEntry(FILE_INFO * fi, byte submode)
{
  int nTop;
  
  if (backstackpushptr == backstackpopptr)
  {
    return 0;
  }

  nTop = prevPtr(backstackpushptr);

  if ((memcmp(&backstack[nTop].fi,fi,sizeof(FILE_INFO)) == 0)
      && (backstack[nTop].submode == submode))
  {
    return 1;
  }

  return 0;
}

int backstackpush(FILE_INFO * fi, byte submode)
{
  if (wouldDuplicateEntry(fi,submode))
  {
    return 0;
  }
  
  memcpy(&backstack[backstackpushptr].fi,fi,sizeof(FILE_INFO));
  backstack[backstackpushptr].submode = submode;

  backstackpushptr = nextPtr(backstackpushptr);

  if (backstackpushptr == backstackpopptr)
  {
    backstackpopptr = nextPtr(backstackpopptr);
  }

  return 1;
}

int backstackpop(FILE_INFO * fi, byte * submode)
{
  if (backstackpushptr != backstackpopptr)
  {
    backstackpushptr = prevPtr(backstackpushptr);
    memcpy(fi,&backstack[backstackpushptr].fi,sizeof(FILE_INFO));
    *submode = backstack[backstackpushptr].submode;
    return 1;
  }

  return 0;
}

