/******************************************************************************
	[Title]		TinyWiki

	Date: 2003-04-02
	Author: J. Steingraeber

	Released to the Public Domain
******************************************************************************/
#include "modeldef.h"

#include "backstack.h"

word callermode;

byte txtbf[2049];

#define MAXWIKILEN 100
char wikiwiki[MAXWIKILEN];
char searchtitle[MAXWIKILEN];
  
FILE_INFO fi;
FILE_BUF fb;

int csize[256];

const TCHTBL FAR TchActionKeyS660[] = {
  {
    0,0,0,0,
    ACT_REPEAT | ACT_MAKE, OBJ_LPSW_UP, 0x00
  }
  ,
  {
    0,0,0,0,
    ACT_MAKE, OBJ_LPSW_PUSH, 0x00
  }
  ,
  {
    0,0,0,0,
    ACT_REPEAT | ACT_MAKE, OBJ_LPSW_DOWN, 0x00
  }
  ,
  {
    0,0,0,0,
    ACT_REPEAT | ACT_MAKE, OBJ_LPSW_LEFT, 0x00
  }
  ,
  {
    0,0,0,0,
    ACT_REPEAT | ACT_MAKE, OBJ_LPSW_RIGHT, 0x00
  }
  ,
  {
        /* NULL end of table marker */
    0,0,0,0,0,0,0
  }
};

#define OBJ_ICON_BACK 0x8001
#define OBJ_ICON_EDIT 0x8002
#define OBJ_TXTAREA 0x8003
#define OBJ_SCROLLBAR 0x8004

const byte FAR backicon[] = {GSIZE(40, 5),
                             0x00, 0x0E, 0x00, 0x40, 0x00,
                             0x00, 0x09, 0x00, 0x40, 0x00,
                             0x00, 0x0E, 0x37, 0x50, 0x00,
                             0x00, 0x09, 0x54, 0x60, 0x00,
                             0xFF, 0xEE, 0x37, 0x57, 0xFF,
};

const word FAR wDummy = 0; /* force word alignment */
const byte FAR editicon[] = {GSIZE(40, 5),
                             0x00, 0x03, 0x8A, 0x80, 0x00,
                             0x00, 0x02, 0x08, 0x80, 0x00,
                             0x00, 0x03, 0x3A, 0xC0, 0x00,
                             0x00, 0x02, 0x2A, 0x80, 0x00,
                             0xFF, 0xFB, 0xBA, 0xDF, 0xFF,
};

const TCHTBL FAR TchToolBar [] = {
  {0, 0, 39, 5, ACT_ICON, OBJ_ICON_BACK, 0x0000}, /* Icon1 */
  {40, 0, 79, 5, ACT_ICON, OBJ_ICON_EDIT, 0x0000}, /* Icon2 */
  {0,0,0,0,0,0,0}
};

const T_ICON FAR TiconBack = {(TCHTBL FAR*)&TchToolBar[0],(byte FAR*)backicon,NULL,0x00};
const T_ICON FAR TiconEdit = {(TCHTBL FAR*)&TchToolBar[1],(byte FAR*)editicon,NULL,0x00};

/*** Designation of text area ***/
#define M_ST_X 0 /* st_x */
#define M_ST_Y 6 /* st_y, must be at least 1 */
#define M_ED_X 147 /* ed_x */
#define M_IT_Y 9 /* it_y */
#define M_MAXG ((160 - M_ST_Y) / M_IT_Y) /* MAXGYO */
#define SCR__Y M_ST_Y /* Scroll bar position */
#define SCR__SIZE M_IT_Y*M_MAXG /* Height of scroll bar */
#define SCR__DSP M_MAXG /* Number of display records */
TCHTBL DdspTch[3]; /* Text touch area table */

TCHSTS tsts;
TXTP m_dd_tp; /* Declaration of structure for text input */
T_SCR_POS m_dd_scb; /* Scroll bar position information */

isWikiWiki(char * p, int len)
{
  int i;
  int maxlen;
  
  wikiwiki[0] = 0;

  if ((p == NULL)
      || (len <= 0))
  {
    return 0;
  }

  if (len >= sizeof(wikiwiki))
  {
    maxlen = sizeof(wikiwiki)-1;
  }
  else
  {
    maxlen = len;
  }
  
  memcpy(wikiwiki,p,maxlen);

  wikiwiki[maxlen] = 0;

  if ((wikiwiki[0] < 'A') || (wikiwiki[0] > 'Z'))
  {
    return 0;
  }

  if ((wikiwiki[1] < 'a') || (wikiwiki[1] > 'z'))
  {
    return 0;
  }
  
  i = 1;
  while ((i < maxlen)
         && ((wikiwiki[i] < 'A') || (wikiwiki[i] > 'Z')))
  {
    ++i;
  }

  if (i == maxlen)
  {
    return 0;
  }

  return 1;
}

/* search for Memo starting with title */
int searchMemo(char * title, word fp)
{
  memset(&fi,0,sizeof(fi));
  memset(&fb,0,sizeof(fb));
  
  fb.fsb_main_entry_ = 0x30; /* MEMO */;
  fb.fsb_sub_entry_ = 0x00; /* all sub modes */
  fb.fsb_scrt_info_ = FILE_OPEN_MODE;
  fi.fp = fp; /*0xFFFF; /* initial search */
  fi.kind = FILE_KIND_TEXT;
  
  while (LibFileFindNext(&fb,&fi,0x00)
         && LibFileRead(&fb,&fi))
  {
    if (title == NULL)
    {
      return 1;
    }
    
    if ((strncmp((char*)fb.fbuf.text.fsb_text_buf_,title,strlen(title)) == 0)
        && (fb.fbuf.text.fsb_text_buf_[strlen(title)] <= 0x20))
    {
      return 1;
    }
    fb.fsb_sub_entry_ = 0x00; /* all sub modes */
  }
  
  return 0;
}

int findWord(char * hay, char * needle)
{
  int i = 0;
  int len = strlen(needle);
  int k = 0;

  while (hay[i])
  {
    if (hay[i] == needle[0])
    {
      k = 0;
      while ((k < len) && (hay[i+k] == needle[k]))
      {
        ++k;
      }

      if (k == len)
      {
        return 1;
      }
    }
    ++i;
  }

  return 0;
}

int searchRef(char * title)
{
  strcpy((char*)txtbf,"References of ");
  strcat((char*)txtbf,title);
  strcat((char*)txtbf,"\r");

  fi.fp = 0xFFFF;
  while (searchMemo(NULL,fi.fp))
  {
    int len = 0;

    while (fb.fbuf.text.fsb_text_buf_[len] > 0x20)
    {
      ++len;
    }

    if ((isWikiWiki((char*)fb.fbuf.text.fsb_text_buf_,len))
        && (strncmp((char*)fb.fbuf.text.fsb_text_buf_,title,strlen(title)) != 0)
        && findWord((char*)&fb.fbuf.text.fsb_text_buf_[1],title))
    {
      if ((strlen((char*)txtbf) + len + 2) >= sizeof(txtbf))
      {
        return 1;
      }
      
      strcat((char*)txtbf,"\r");
      txtbf[strlen((char*)txtbf)+len] = 0;
      memcpy(&txtbf[strlen((char*)txtbf)],fb.fbuf.text.fsb_text_buf_,len);
    }
  }

  memcpy(fb.fbuf.text.fsb_text_buf_,txtbf,sizeof(txtbf));
  
  return 1;
}

void matchWord(TXTP * txtp,int * sp, int * ep)
{
  if (sp != NULL)
  {
    *sp = 0;
  }

  if (ep != NULL)
  {
    *ep = 0;
  }
  
  if ((txtp == NULL) || (sp == NULL) | (ep == NULL))
  {
    return;
  }

  *sp = txtp->sr_ep; /* start with end pointer, too */
  *ep = txtp->sr_ep;
  
  while ((*sp)
         && (txtp->txbf[*sp-1] >= 'A')
         && (txtp->txbf[*sp-1] < 0xFE))
  {
    --*sp;
  }

  *ep = *sp;
  while ((txtp->txbf[*ep] >= 'A')
         && (txtp->txbf[*ep] < 0xFE))
  {
    ++*ep;
  }
}

int drawIconIsClicked(TCHSTS * tsts, T_ICON FAR * icon)
{
  if (tsts->obj == icon->tch->obj)
  {
    if (tsts->act & (ACT_MAKE | ACT_MOVE_IN))
    {
      LibIconPrintR(icon);
    }
    else if (tsts->act & (ACT_MOVE_OUT | ACT_BREAK_IN))
    {
      LibIconPrint(icon);
    }

    if (tsts->act & ACT_BREAK_IN)
    {
      return 1;
    }
  }

  return 0;
}

void callEditor(word cursor)
{
#ifdef __HITACHI__
  char * p = (char *)0x8C001922;
  void * modeinfo;
#else  
  char far * p = MK_FP(0x1840,0x08BE);
  word myseg, myoff;
#endif
  byte name[MAX_FILENAMELEN];
  byte sub_mode;
  byte status;
  word mymode, mystatus;
  word mode;
  int i, n;

  mode = IW_MMEMO;
  
#ifdef __HITACHI__
  LibGetMode(&mymode,&mystatus,&modeinfo);
#else
  LibGetMode(&mymode,&mystatus,&myseg,&myoff);
#endif
  
  ((word FAR*)p)[0] = 0x2708;
  p -= 2;
  ((word FAR*)p)[0] = mymode;
  p -= 2;

  sprintf((char*)name,"%d",fb.fsb_sub_entry_);
  n = strlen((char*)name);
  p -= n - 1;
  
  for (i=0;i<n;++i)
  {
    p[n-i] = name[i];
  }
  p[0] = 0xFE;
  --p;

  i = 0;
  while (fb.fbuf.text.fsb_text_buf_[i] > 0x20)
  {
    p[0] = fb.fbuf.text.fsb_text_buf_[i];
    --p;
    ++i;
  }
  p[0] = 0xFE;
  --p;

  sprintf((char*)name,"%d",cursor);
  n = strlen((char*)name);
  p -= n;
  
  for (i=0;i<n;++i)
  {
    p[n-i] = name[i];
  }
  p[0] = 0;
  
  status = 0;

  if (callermode)
  {
    status = 0xF0;
    mode = callermode;
    sub_mode = 0;
  }
  else
  {
    for (sub_mode=1;sub_mode<64;++sub_mode)
    {
      if ((LibGetProgramName(name,IB_MADDIN,sub_mode))
          && (strncmp((char*)name,"IEdit",5) == 0))
      {
        status = 0xF0;
        break;
      }
    }
  
    if (status) /* IEdit was found */
    {
      mode = IW_MADDIN;
    }
    else /* use Memo as editor */
    {
      sub_mode = 0;
          /* help Memo to start with current category */
#ifdef __HITACHI__
      *((char *)0x8C001DF6) = fb.fsb_sub_entry_;
#else
      *((char far *)MK_FP(0x1840,0x00B2)) = fb.fsb_sub_entry_;
#endif    
    }
  }
  
  LibModeJump(mode + sub_mode,status);
}

char * getStartpage()
{
#ifdef __HITACHI__
  char * p = (char *)0x8C001922;
  void * modeinfo;
#else  
  char far * p = MK_FP(0x1840,0x08BE);
  word myseg, myoff;
#endif
  word mymode, mystatus;
  int i = 0;
  
#ifdef __HITACHI__
  LibGetMode(&mymode,&mystatus,&modeinfo);
#else  
  LibGetMode(&mymode,&mystatus,&myseg,&myoff);
#endif

  if (((mystatus & 0xF0) == 0) /* not started via modecall */
      || (((word FAR*)p)[0] != 0x2708)) /* magic number not present */
  {
    return NULL;
  }
  
  p -= 2;
  callermode = ((word FAR*)p)[0];
  p -= 2;

      /* skip categoryy */
  while (*p && (*p != 0xFE))
  {
    --p;
  }

  if (*p == 0xFE)
  {
    --p;
  }

      /* get name of memo */
  while (*p && (*p != 0xFE))
  {
    wikiwiki[i] = *p;
    ++i;
    --p;
  }
  wikiwiki[i] = 0;

  return wikiwiki;
}

void insertWordWrap(TXTP * txtp)
{
  word idx;
  int xleft = txtp->st_x;
  int xright = txtp->ed_x;
  byte c;
  int x = xleft;
  word idxbreak = (word)-1;
  
  for (c = 0x20; c < 0xFE ; ++c)
  {
    csize[c] = LibGetProFontSize(txtp->font,c);
  }

  idx = 0;

  c = txtp->txbf[idx];
  
  while (c)
  {
    if ((c == 0x0D) || (c == 0xFE))
    {
      x = xleft;
      idxbreak = (word)-1;
    }
    
    if (c >= 0x20)
    {
      x += csize[c];
    }

    if ((x >= xright) && (idxbreak != (word)-1))
    {
      txtp->txbf[idxbreak] = 0xFE;
      x = xleft;
      idx = idxbreak + 1;
      idxbreak = (word)-1;
      c = txtp->txbf[idx];
      continue;
    }

    if (c == 0x20)
    {
      idxbreak = idx;
    }

    ++idx;
    c = txtp->txbf[idx];
  }
}

int findXY(TXTP * txtp, word pos, int * x, int * y)
{
  int i;
  word w;
  
  for (i=1;i<22;++i)
  {
    if (txtp->dlmp[i] > pos)
    {
      break;
    }
  }

  if ((i == 1) || (i == 22))
  {
    return 0;
  }

  w = txtp->dlmp[i-1];
  if (w == 0xFFFF)
  {
    return 0;
  }

  *x = 0;
  while (w != pos)
  {
        /* LibGetProStrSize() ignores white space at end of string :-( */
    *x += LibGetProFontSize(txtp->font,txtp->txbf[w]);
    ++w;
  }
  *x += txtp->st_x;
  *y = (i-2) * txtp->it_y + txtp->st_y;

  return 1;
}

void emuTouch(TXTP * txtp, word obj, int x1, int y1, int x2, int y2)
{
  TCHSTS tsts;
  
  memset(&tsts,0,sizeof(TCHSTS));
  tsts.obj = obj;
  tsts.act = ACT_MAKE;
  tsts.x = x1;
  tsts.y = y1;
  LibTxtDspS(txtp,&tsts);
  LibTxtDspC(txtp);

  tsts.x = x2;
  tsts.y = y2;

  if ((x1 != x2) || (y1 != y2))
  {
    tsts.act = ACT_DOWN_IN;
    LibTxtDspS(txtp,&tsts);  
    LibTxtDspC(txtp);
  }

  tsts.act = ACT_BREAK;
  LibTxtDspS(txtp,&tsts);  
/*   LibTxtDspC(txtp); */
}

char * showMemo()
{
  int sp, ep;
  
  memcpy(txtbf,fb.fbuf.text.fsb_text_buf_,sizeof(txtbf));
    
      /** Text input setting **/
  m_dd_tp.st_x = M_ST_X; /* Start coordinate (X) of text display */
  m_dd_tp.st_y = M_ST_Y; /* Start coordinate (Y) of text display */
  m_dd_tp.ed_x = M_ED_X; /* End coordinate (X) of text display */
  m_dd_tp.it_y = M_IT_Y; /* Text display line spacing (Y) */
  m_dd_tp.MAXGYO = M_MAXG; /* Number of text display lines */
  m_dd_tp.font = IB_PFONT1; /* Display font type */
  m_dd_tp.csen = FALSE; /* Cursor display disabled */
  m_dd_tp.rtnen = FALSE; /*FALSE; /* CR code display disabled */
  m_dd_tp.maxmj = 2048; /* Maximum number of allowable input characters. */
  m_dd_tp.txbf = txtbf; /* Designation of text buffer address */
  m_dd_tp.txtobj = OBJ_TXTAREA; /* Object when text area is touched. */
  m_dd_tp.sbrobj = OBJ_SCROLLBAR; /* Object when the scroll bar is touched*/
  m_dd_tp.tchtb = DdspTch; /* Pointer for text scroll bar area */
  m_dd_tp.s_pos = &m_dd_scb; /* Pointer for text and scroll bar information */

  insertWordWrap(&m_dd_tp);
  
      /*** Text touch area setting ***/
  LibTchStackClr();
  LibTchStackPush( NULL );
  LibTchStackPush((TCHTBL FAR*)TchToolBar);
  LibTchStackPush(TchHardIcon); /* Hard icon touch table */
  LibTchStackPush((TCHTBL FAR*)TchActionKeyS660);
  LibTxtDspInit(&m_dd_tp); /* Initialization of text input */
  LibTxtTchSet(&m_dd_tp); /* Text touch area PUSH */
  
  LibClrDisp();
  LibLine(0,M_ST_Y-2,160,1,1);
  
  LibIconPrint(&TiconBack);
  LibIconPrint(&TiconEdit);
  
      /*** Data display loop ***/
  while(1)
  {
    LibTxtDspC(&m_dd_tp);
    LibPutDisp();

    LibTchWait( &tsts ); /* Waiting for touch */

    if ((tsts.obj == OBJ_TXTAREA)
        && (tsts.act & ACT_BREAK))
    {
      matchWord(&m_dd_tp,&sp,&ep);
      if (isWikiWiki((char*)&txtbf[sp],ep-sp))
      {
        if (sp == 0)
        {
          strcpy(searchtitle,wikiwiki);
          if (searchRef(searchtitle))
          {
            return (char*)-2;
          }
        }
          
        if (searchMemo(wikiwiki,0xFFFF))
        {
          return wikiwiki;
        }
      }
    }

    if (tsts.obj == OBJ_HIC_ESC)
    {
      break;
    }

    if (tsts.obj == OBJ_LPSW_LEFT)
    {
      emuTouch(&m_dd_tp,OBJ_SCROLLBAR,155,10,155,10);
      continue;
    }
    
    if (tsts.obj == OBJ_LPSW_UP)
    {
      emuTouch(&m_dd_tp,OBJ_SCROLLBAR,155,14,155,14);
      continue;
    }
    
    if (tsts.obj == OBJ_LPSW_DOWN)
    {
      emuTouch(&m_dd_tp,OBJ_SCROLLBAR,155,149,155,149);
      continue;
    }
    
    if (tsts.obj == OBJ_LPSW_RIGHT)
    {
      emuTouch(&m_dd_tp,OBJ_SCROLLBAR,155,155,155,155);
      continue;
    }
    
    if (drawIconIsClicked(&tsts,(T_ICON FAR*)&TiconBack))
    {
      return (char*)-1;
    }
    
    if (drawIconIsClicked(&tsts,(T_ICON FAR*)&TiconEdit))
    {
      callEditor(m_dd_tp.cp);
    }
      
    LibTxtDspS(&m_dd_tp,&tsts);
  }
  LibTchStackClr();
  
  return NULL;
}

void main()
{
  char frontpage[] = "FrontPage";
  char * title;
  FILE_INFO myfi;
  byte submode;
  int isPseudoText = 0;
  int returnToCaller = 0;
  
  title = getStartpage();
  
  if (title == NULL)
  {
    title = frontpage;
    callermode = 0;
  }
  else
  {
    returnToCaller = 1;
  }
  
  if (!searchMemo(title,0xFFFF))
  {
    fb.fsb_sub_entry_ = 0;
    strcpy(fb.fbuf.text.fsb_text_buf_,"FrontPage\r\r... could not be found.\rPlease create a memo with the word \"FrontPage\" in the first line.");
    showMemo();
    LibJumpMenu();
  }
  
  while (title != NULL)
  {
    if (title == (char*)-2)
    {
      isPseudoText = 1;
    }
    else
    {
      memcpy(&myfi,&fi,sizeof(FILE_INFO));
      submode = fb.fsb_sub_entry_;
      isPseudoText = 0;
    }
    
    title = showMemo();

    if (title == (char*)-1)
    {
      if (isPseudoText)
      {
        memcpy(&fi,&myfi,sizeof(FILE_INFO));
        fb.fsb_sub_entry_ = submode;
      }
      else
      {
        backstackpop(&fi,&fb.fsb_sub_entry_);
      }
      if (fb.fsb_sub_entry_)
      {
        LibFileRead(&fb,&fi);
      }
    }
    else if ((title == (char*)-2)
             || ((title != NULL)
                 && (!isPseudoText)))
    {
      backstackpush(&myfi,submode);
    }
    
  }

  if (returnToCaller && !isPseudoText)
  {
    callEditor(0);
  }
  
  LibJumpMenu();
}
